#!/usr/bin/perl
#################################### Modules ##########################################
#is required for perl2exe compiling
#perl2exe_include "attributes.pm"
#perl2exe_include "Log/Log4perl/Appender/Screen.pm"
#perl2exe_include "Log/Log4perl/Appender/File.pm"
#perl2exe_include "Net/SNMP/Transport/IPv4.pm"
#
require v5.8.6;

use strict;
use warnings;
use feature qw(switch);
use Time::HiRes qw(gettimeofday);
use Log::Log4perl qw(:easy);
use Spreadsheet::WriteExcel;
use ERS_MIB 0.07;
use ERS_data 0.07;

#################################### init Variables ###################################
my $logger;										# LOG object
my %cfg;										# all run parameter
my @hosts;										# load IP list for work
my %data;										# contain all collected information
my $version	= '0.0.9';
my @report;
my $processTime;
my @days	= qw(dummy Monday Tuesday Wednesday Thursday Friday Saturday Sunday);
$| = 1;

#######################################################################################
#################################### Main #############################################
#######################################################################################

initArguments();

start();

writeFile();

#######################################################################################
################################ Subroutines ##########################################
#######################################################################################
#initArguments ();
sub initArguments{
  my $startUp = $0;
  $startUp =~ /[^\\\/]+\.[\w]+$/;
	$startUp = $&;
	my ($syntax) = <<EOFinfo;
Avaya Energy Saver scrip, version $version

Usage: $startUp -h <ip/file> -c <comunity> [options]

Parameter:
  -h <ip/file>    IP or file (a listof IPs) of ERS IP address
  -c <comunity>   SNMPv2 read comunity
  -f <file>       name of the Excel file
  
Option:
  -t <time>       SNMP timeout in seconds (default 5)
  -r <number>     SNMP retries (default 1)
  -p <number>     kWh price in cents
  -u <text>       currency (default EUR)
  -debug
  -trace
  
Examples:
	$startUp -h ip_hosts.txt -c public -f result.xls
	$startUp -h ip_hosts.txt -c public -f result.xls -p 26.73 -u USD
	$startUp -h ip_hosts.txt -c public -f result.xls -t 6 -r 3
	$startUp -h ip_hosts.txt -c public -f result.xls -debug
				
EOFinfo

	if(!defined($ARGV[0])){
		die $syntax;
	}else{	
		$cfg{'logLevel'} = 0;
		my $count = 0;
		foreach my $in (@ARGV){
			$cfg{'iplist'}		= $ARGV[$count + 1]	if($in =~ /^-h$/);
			$cfg{'snmpComunity'}	= $ARGV[$count + 1]	if($in =~ /^-c$/);
			$cfg{'snmpTimeout'}	= $ARGV[$count + 1]	if($in =~ /^-t$/);
			$cfg{'snmpRetries'}	= $ARGV[$count + 1]	if($in =~ /^-r$/);
			$cfg{'price'}		= $ARGV[$count + 1]	if($in =~ /^-p$/);
			$cfg{'unit'}		= $ARGV[$count + 1]	if($in =~ /^-u$/);
			$cfg{'file'}		= $ARGV[$count + 1]	if($in =~ /^-f$/);
			$cfg{'logLevel'}	= 1			if($in =~ /^-debug$/);
			$cfg{'logLevel'}	= 2			if($in =~ /^-trace$/);
			$count++;
		}
		
		setupLogFacility($cfg{'logLevel'});
		
		$logger->debug("Script Version: $version");
		
		if(!$cfg{'iplist'} or !$cfg{'snmpComunity'} or !$cfg{'file'}){
			$logger->fatal("missing required parameter");
			print "\n\n$syntax";
			exit 1;
		}
		
		if($cfg{'snmpTimeout'} and ($cfg{'snmpTimeout'} !~ /^\d+$/ or $cfg{'snmpTimeout'} > 60)){
			$logger->fatal("SNMP timeout '$cfg{'snmpTimeout'}' not valide, only number between 1-60 allowed.");
			exit 1;
		}
		
		if($cfg{'snmpRetries'} and ($cfg{'snmpRetries'} !~ /^\d+$/ or $cfg{'snmpRetries'} > 20)){
			$logger->fatal("SNMP retries '$cfg{'snmpRetries'}' not valide, only number between 1-5 allowed.");
			exit 1;
		}
		
		if($cfg{'price'} and $cfg{'price'} !~ /^\d+\.\d+$/){
			$logger->fatal("price per kWh '$cfg{'snmpRetries'}' not valide, use 26.73 cents.");
			exit 1;
		}
		
		parsertr(\@hosts, $cfg{'iplist'});
		
		$cfg{'allHostCount'} = @hosts;
		
		return;
	}
}

#######################################################################################
#start ();
sub start{
	
	$logger->debug("creating data objects");
	foreach my $ip (@hosts){
		$data{$ip} = ERS_data->new($ip, $logger);
	}
	
	$logger->debug("creating SNMP objects");
	$cfg{'snmp'} = ERS_MIB->new($logger, $cfg{'snmpComunity'}, $cfg{'snmpTimeout'}, $cfg{'snmpRetries'});
	if(!$cfg{'snmp'}){
		$logger->error("by building SNMP object");
		exit 1;
	}
	
	$logger->debug("start SNMP processing");
	my $startTime = gettimeofday;
	
	my @hostCopy = @hosts;
	foreach my $job (qw(determineOID getUnitCount getPortCount getEnergySaverEnabled interfaceEnergySaverEnabled energyUnitSavings getEnergySaverSchedule)){
		$cfg{'snmp'}->$job(\@hostCopy, \%data);
	}
	
	$processTime = sprintf "%.3f", (gettimeofday - $startTime);
	$logger->debug("$processTime seconds consumed for collecting");
}

#######################################################################################
#writeFile()
sub writeFile {
	my %excel;
	my $switchCount = 0;
	my $unitCount = 0;
	my $portCount = 0;
	my $energySaverEnabledCount = 0;
	my $savingsCount = 0;
	my $badHostCount = 0;
	my %lostHostCount;
	my $goodHostCount = 0;
	my ($sec, $min, $hour, $day, $month, $year) = (localtime)[0..5];
	$sec = '0'.$sec if length($sec) < 2;
	$min = '0'.$min if length($min) < 2;
	$hour = '0'.$hour if length($hour) < 2;
	$day = '0'.$day if length($day) < 2;
	$month++;
	$month = '0'.$month if length($month) < 2;
	$year = $year + 1900;
	my ($file, $extention) = split /\./, $cfg{'file'};
	$file = $file."_$year-$month-$day".'_'."$hour-$min-$sec.".$extention;
	
	$logger->debug("creating Excel file '$file'");
	
	$excel{'workbook'}  = Spreadsheet::WriteExcel->new($file)
		or die "\n";
	my $header = $excel{'workbook'}->add_format();
        $header->set_bold();
        $header->set_color('blue');
        $header->set_align('center');
	my $banner = $excel{'workbook'}->add_format();
        $banner->set_bold();
        $banner->set_size(12);
	$banner->set_color('red');
        $banner->set_align('center');
	my $left = $excel{'workbook'}->add_format();
	$left->set_align('left');
	my $center = $excel{'workbook'}->add_format();
	$center->set_align('center');
	my $right = $excel{'workbook'}->add_format();
	$right->set_bold();
	$right->set_align('right');
	my $good = $excel{'workbook'}->add_format();
        $good->set_bold();
        $good->set_color('green');
	my $bad = $excel{'workbook'}->add_format();
        $bad->set_bold();
        $bad->set_color('red');
	
	$logger->trace("create Summary sheet part I");
	$excel{'Summary'} = $excel{'workbook'}->add_worksheet('Summary');
	$excel{'Summary'}->add_write_handler(qr[\w], \&store_string_widths);
	$excel{'Summary'}->write(0, 0, "Summary", $header);
	$excel{'Summary'}->write(0, 2, "Avaya Energy Saver", $banner);
	$excel{'Summary'}->write(2, 0, "Time:", $right);
	$excel{'Summary'}->write(2, 1, "$hour:$min:$sec");
	$excel{'Summary'}->write(2, 2, "$day-$month-$year");
	$excel{'Summary'}->write(3, 0, "Scrip version:", $right);
	$excel{'Summary'}->write(3, 1, $version);
	$excel{'Summary'}->write(4, 0, 'Script runtime', $right);
	$excel{'Summary'}->write(4, 1, $processTime);
	$excel{'Summary'}->write(4, 2, "seconds");
	
	$logger->trace("create Host sheet");
	$excel{'Hosts'} = $excel{'workbook'}->add_worksheet('Hosts');
	$excel{'Hosts'}->add_write_handler(qr[\w], \&store_string_widths);
	$excel{'Hosts'}->write(0, 0,  "IP", $header);
	$excel{'Hosts'}->write(0, 1,  "Status", $header);
	$excel{'Hosts'}->write(0, 2,  "Type", $header);
	$excel{'Hosts'}->write(0, 3,  "Name", $header);
	$excel{'Hosts'}->write(0, 4,  "Software release", $header);
	$excel{'Hosts'}->write(0, 5,  "Description", $header);
	my $row = 0;
	foreach my $ip (@hosts){
		my $column = 0;
		$row++;
		$excel{'Hosts'}->write($row, $column,  $ip);
		if($data{$ip}->getValue('Status')){
			foreach my $statement (qw(Status Type Name Software Description)){
				$column++;
				if($statement eq 'Status'){
					my $format = $good;
					$format = $bad  if $data{$ip}->getValue($statement) ne 'up';
					$excel{'Hosts'}->write($row, $column,  $data{$ip}->getValue($statement), $format);
				}else{
					$excel{'Hosts'}->write($row, $column,  $data{$ip}->getValue($statement));
				}
			}
		}else{
			$excel{'Hosts'}->write($row, 1,  'unknow');
		}
	}
	$logger->debug(@hosts." hosts in 'Host' sheet");
	
	$logger->trace("create Details sheet");
	$excel{'Details'} = $excel{'workbook'}->add_worksheet('Details');
	$excel{'Details'}->add_write_handler(qr[\w], \&store_string_widths);
	$excel{'Details'}->write(0, 0,  "IP", $header);
	$excel{'Details'}->write(0, 1,  "UnitCount", $header);
	$excel{'Details'}->write(0, 2,  "EnergySaverEnabled", $header);
	$excel{'Details'}->write(0, 3,  "PortCount", $header);
	$excel{'Details'}->write(0, 4,  "interfaceEnergySaverEnabled", $header);
	$excel{'Details'}->write(0, 5,  "EnergyUnitSavings", $header);
	
	my %hostSavings;
	$row = 0;
	foreach my $ip (@hosts){
		if($data{$ip}->getValue('Status') ne 'up'){
			$logger->trace("[$ip] host ignored");
			$badHostCount++;
			next;
		}else{
			$goodHostCount++;
		}
		
		my $fromatIp;
		my $column = 0;
		$row++;
		my $makerBadIp = 0;
		foreach my $statement (qw(UnitCount EnergySaverEnabled PortCount interfaceEnergySaverEnabled EnergyUnitSavings)){
			$column++;
			my $formatValue = '';
			my $value = $data{$ip}->getValue($statement);
			if(!defined($value)){
				$value = 0;
				$fromatIp = $bad;
				$formatValue = $bad;
				$makerBadIp = 1;
			}
			$excel{'Details'}->write($row, $column,  $value, $formatValue);
			given ($statement){
				when('UnitCount') { $unitCount = $unitCount + $value; }
				when('PortCount') { $portCount = $portCount + $value; }
				when('interfaceEnergySaverEnabled') { $energySaverEnabledCount = $energySaverEnabledCount + $value; }
				when('EnergyUnitSavings') { $savingsCount = $savingsCount + $value; $hostSavings{$ip} = $value}
				default {  }
			}
		}
		$excel{'Details'}->write($row, 0,  $ip, $fromatIp);
		$lostHostCount{$ip} = 1 if $makerBadIp;
	}
	$logger->debug(@hosts." hosts [good($goodHostCount)/bad($badHostCount)] in 'Details' sheet");
	
	$logger->trace("create Schedule sheet");
	$excel{'Schedule'} = $excel{'workbook'}->add_worksheet('Schedule');
	$excel{'Schedule'}->add_write_handler(qr[\w], \&store_string_widths);
	$excel{'Schedule'}->write(0, 0,  "IP", $header);
	$excel{'Schedule'}->write(0, 1,  "minutes activated", $header);
	$excel{'Schedule'}->write(0, 2,  "Schedule (deactivate)", $header);
	
	$row = 0;
	my $scheduleCount = 0;
	foreach my $ip (@hosts){
		if($data{$ip}->getValue('Status') ne 'up'){
			$logger->trace("[$ip] host ignored");
			next;
		}
		my $column = 1;
		$row++;
		if($data{$ip}->getValue('EnergySaverSchedule')){
			$excel{'Schedule'}->write($row, 0,  $ip);
			my %buffer;
			my $count = 0;
			foreach my $line (@{$data{$ip}->getValue('EnergySaverSchedule')}){		# prep to sort days later on
				if($line->{'day'} and $line->{'hour'} and $line->{'minute'}){
					push(@{$buffer{$days[$line->{'day'}]}}, $line);
				}else{
					$buffer{$count} = 'corrupt data';
				}
				$count++;
			}
			
			my $dayCount = 0;
			my $activate = 0;
			foreach my $day (@days){
				if($buffer{$day}){
					$dayCount++;
					my %daySchedule;
					foreach my $line (@{$buffer{$day}}){					# prep time per day
						$daySchedule{$line->{'hour'}.$line->{'minute'}} = $line;
					}
					my $dayString;
					foreach my $row (sort keys %daySchedule){
						my $line = $daySchedule{$row};
						$dayString = $days[$line->{'day'}] if !$dayString;
						$dayString .= " $line->{'hour'}:$line->{'minute'} -";
						if($line->{'action'} eq 'deactivate'){
							$activate = $activate + ($line->{'hour'} * 60) + $line->{'minute'};
						}else{
							$activate = $activate + (1440 - ($line->{'hour'} * 60) + $line->{'minute'});
						}
					}
					if($dayString){
						$dayString =~ s/-$//;
						$column++;
						$excel{'Schedule'}->write($row, $column, $dayString);
					}
				}
			}
			
			if($activate and $dayCount < 7){
				$activate = $activate + ((7 - $dayCount) * 1440);
			}
			
			$excel{'Schedule'}->write($row, 1, $activate);
			$hostSavings{$ip} = $hostSavings{$ip} * ($activate / 60);
		}else{
			$excel{'Schedule'}->write($row, 0, $ip, $bad);
			$excel{'Schedule'}->write($row, 3, 'data missing');
			$logger->warn("scheduled events missing for host $ip");
			$lostHostCount{$ip} = 1;
		}
	}
	$logger->debug("$scheduleCount/$row hosts have schedule events in 'Schedule' sheet");
	
	$logger->trace("create Summary sheet part II");
	$excel{'Summary'}->write(6, 0, "Host:", $right);
	$excel{'Summary'}->write(6, 1, $cfg{'allHostCount'});
	$excel{'Summary'}->write(6, 2, "total");
	$excel{'Summary'}->write(7, 0, "Host:", $right);
	$excel{'Summary'}->write(7, 1, $goodHostCount);
	$excel{'Summary'}->write(7, 2, "available");
	$excel{'Summary'}->write(8, 0, "Host:", $right);
	$excel{'Summary'}->write(8, 1, $badHostCount);
	$excel{'Summary'}->write(8, 2, "no response");
	my $countHost = 0;
	$countHost++ foreach (keys %lostHostCount);
	$excel{'Summary'}->write(9, 0, "Host:", $right);
	$excel{'Summary'}->write(9, 1, $countHost);
	$excel{'Summary'}->write(9, 2, "lost communication");
	$excel{'Summary'}->write(11, 0, "ERS Unit:", $right);
	$excel{'Summary'}->write(11, 1, $unitCount);
	$excel{'Summary'}->write(12, 0, "ERS Port:", $right);
	$excel{'Summary'}->write(12, 1, $portCount);
	$excel{'Summary'}->write(12, 2, "total");
	$excel{'Summary'}->write(13, 0, "ERS Ports:", $right);
	$excel{'Summary'}->write(13, 1, $energySaverEnabledCount);
	$excel{'Summary'}->write(13, 2, "Energy Saver enabled");
	
	my ($value, $unit) = units($savingsCount);
	$excel{'Summary'}->write(14, 0, "current Energy Savings:", $right);
	$excel{'Summary'}->write(14, 1, $value);
	$excel{'Summary'}->write(14, 2, $unit);
	
	my $pwSavings = 0;
	foreach my $ip (keys %hostSavings){
		$pwSavings =  $pwSavings + $hostSavings{$ip}
	}
	($value, $unit) = units($pwSavings);
	my $paSavings = ($pwSavings / 7) * 365;
	($value, $unit) = units($paSavings, 'force');
	$excel{'Summary'}->write(15, 0, "estimated Energy Savings pa:", $right);
	$excel{'Summary'}->write(15, 1, $value);
	$excel{'Summary'}->write(15, 2, $unit.'h');
	
	if($cfg{'price'}){
		$excel{'Summary'}->write(16, 0, "estimated Cost Savings per pa:", $right);
		$excel{'Summary'}->write(16, 1, (sprintf "%.2f", ($paSavings / 1000) * ($cfg{'price'} / 100)), $good);
		if($cfg{'unit'}){
			$excel{'Summary'}->write(16, 2, $cfg{'unit'}, $good);
		}else{
			$excel{'Summary'}->write(16, 2, 'EUR', $good);
		}
	}
	
	foreach my $sheet (qw(Summary Hosts Details Schedule)){
		autofit_columns($excel{$sheet});
	}
	
	$logger->info("Excel file '$file' created");
}

#######################################################################################
#units()
sub units {
	my $value = shift;
	my $force = shift || 0;
	my @ret;
	
	if($value >= 1000 or $force){
		$ret[0] = sprintf "%.1f", $value / 1000;
		$ret[1] = 'kW';
	}else{
		$ret[0] = sprintf "%.1f", $value;
		$ret[1] = 'W';
	}
	return @ret;
}

#######################################################################################
#autofit_columns()
# Adjust the column widths to fit the longest string in the column.
sub autofit_columns {
	my $worksheet = shift;
	my $col       = 0;
	
	$logger->trace("adjust column width in Excel file");
	
	for my $width (@{$worksheet->{__col_widths}}) {
		$worksheet->set_column($col, $col, $width) if $width;
		$col++;
	}
}

#######################################################################################
#autofit_columns()
sub store_string_widths {
	my $worksheet = shift;
	my $col       = $_[1];
	my $token     = $_[2];
	
	# Ignore not relevant parts
	return if not defined $token;						# undefs.
	return if $token eq '';							# blank cells.
	return if ref $token eq 'ARRAY';					# array refs.
	return if $token =~ /^=/;						# formula
	return if $token =~ /^([+-]?)(?=\d|\.\d)\d*(\.\d*)?([Ee]([+-]?\d+))?$/;	# numbers
	
	# Ignore internal and external hyper links.
	return if $token =~ m{^[fh]tt?ps?://};
	return if $token =~ m{^mailto:};
	return if $token =~ m{^(?:in|ex)ternal:};
	
	my $old_width    = $worksheet->{__col_widths}->[$col];
	my $string_width = string_width($token);
	
	if (not defined $old_width or $string_width > $old_width) {
		$worksheet->{__col_widths}->[$col] = $string_width;
	}
	
	return undef;
}

#######################################################################################
#string_width()
sub string_width {
	
	return 1.1 * length $_[0];
}

#######################################################################################
#parsertr(<"listfile">");		extract a list of IP addresses 
sub parsertr{
	my $refA_hosts = shift;
	my $listfile = shift;
	my @list;
	my $problem = 0;
	
	if($listfile =~ /^(\d{1,3}\.){3}\d{1,3}$/){
		push(@$refA_hosts, $listfile);
	}else{
		open HOSTINFILE,"<$listfile"
			or $problem = 1;
		
		if($problem){
			$logger->fatal("can not read host list file: '$listfile'");
			exit 3;
		}
		
		my @listin = <HOSTINFILE>;
		close HOSTINFILE;
		chomp @listin;
		
		my $count = 0;
		foreach(@listin){
			while($listin[$count] =~ /[\n\t\s]|(#.*$)/){
				$listin[$count] =~ s/[\n\t\s]|(#.*$)//;
			}
			
			if($listin[$count] ne "" and  $listin[$count] =~ /^(\d{1,3}\.){3}\d{1,3}/){
				my $ip = $&;
				my $pwd = $';
				
				if($pwd){
					$cfg{'specificPwd', $ip} = $';
					$cfg{'specificPwd', $ip} =~ s/^;|,//;
				}
				
				my $signal = 0;
				foreach(@$refA_hosts){
					if($ip eq $_){
						$signal = 1;
						last;
					}
				}
				
				if($signal){
					$logger->warn("ignore dublicate host '$ip' in file '$listfile'");
				}else{
					$logger->trace("add IP $ip to process list");
					push(@$refA_hosts, $ip);
				}
			}
			$count++;
		}
	}
	
	my $length = @$refA_hosts;
	if($length){
		$logger->debug("register $length hosts");
	}else{
		$logger->fatal("no IP addresses found in '$listfile'");
		exit 1;
	}
}

#######################################################################################
#setupLogFacility()	
sub setupLogFacility {
	my $logLevel = shift || 0;
	my @logLevels = qw(INFO DEBUG TRACE);
	my %log_conf = ('log4perl.rootLogger'					=> $logLevels[$logLevel].', Screen, Logfile',
			'log4perl.appender.Screen'				=> 'Log::Log4perl::Appender::Screen',
			'log4perl.appender.Screen.layout'			=> 'Log::Log4perl::Layout::PatternLayout',
			'log4perl.appender.Screen.layout.ConversionPattern'	=> '%-7.7r %-7.7p %-70.70m [%-4.4L %-12.12M{2} %-12.12F{1}]%n',
			'log4perl.appender.Logfile'				=> 'Log::Log4perl::Appender::File',
			'log4perl.appender.Logfile.filename'			=> 'ERS-AES.log',
			'log4perl.appender.Logfile.layout'			=> 'Log::Log4perl::Layout::PatternLayout',
			'log4perl.appender.Logfile.layout.ConversionPattern'	=> '%-7.7r %-7.7p %-100.100m [%-4.4L %-30.30M{2} %-12.12F{1}]%n',
			);
	
	if (!defined($logger)){
		$logger = Log::Log4perl->get_logger();
		Log::Log4perl::init(\%log_conf);
		$logger->debug("log facility start with [$logLevels[$logLevel]]");
	}else{
		Log::Log4perl::init(\%log_conf);
		$logger->info("log facility updated [$logLevels[$logLevel]]");
	}
}
