#######################################################################################
###### This package handles the interaction with the ERS switch via SNMP     ##########
#######################################################################################
package ERS_MIB;
our $VERSION = '0.07';

use strict;
use Net::SNMP qw(:snmp);
use Time::HiRes qw(gettimeofday);
use MIBstuff 0.07;
use ERS_data 0.07;

#################################### init Variables ###################################
my $maxBulkRepition = 51;
my $timeout = 5;
my $retries = 1;
my $startTime = 0;

#######################################################################################
# new
sub new {
	my $class = shift;
	my $logger = shift;
	my $snmpComunity = shift;
	my $snmpTimeout = shift || $timeout;
	my $snmpRetries = shift || $retries;
	my $self = {};
	
	bless($self, $class);
	
	$self->{_name}		= '';
	$self->{_logger}	= $logger;
	$self->{_comunity}	= $snmpComunity;
	$self->{_timeout}	= $snmpTimeout;
	$self->{_retries}	= $snmpRetries;
	$self->{_ERRMSG}	= '';
	$self->{'oids'}		= [];
	
	$self->{_logger}->debug("SNMP settings: timeout=$snmpTimeout, retries=$snmpRetries");
	
	return $self;
}

#######################################################################################
#DESTROY();
sub DESTROY {
	my $self = shift;
	
}

#######################################################################################
#getErrorMsg();
sub getErrorMsg {
	my $self = shift;
	
	return $self->{_ERRMSG};
}

#######################################################################################
#performSNMP(job, action, Host, MIBname, callback)
sub performSNMP {
	my $self 	= shift;
	my $action 	= shift;
	my $callBack	= shift;
	my $refA_hosts	= shift;
	my $refA_MIBs	= shift;
	my $refO_data	= shift;
	my $maxRepetitions = shift || 10;						# only for get_bulk_request
	
	$self->{_logger}->trace("run SNMP $action $callBack for ".@{$refA_hosts}." hosts");
	
	$callBack .= '_cb';
	my %params	= (	-varbindlist	=> $refA_MIBs,
				-callback	=> [ \&$callBack, $self->{_logger}, $refA_hosts, $refO_data ],
			);
	
	if($action eq 'get_bulk_request'){
		$params{'-maxrepetitions'} = $maxRepetitions;
	}
	
	$self->{_logger}->trace("execute  SNMP $action ".@{$refA_MIBs}."/$maxRepetitions");
	
	$startTime = gettimeofday;
	foreach my $ip (@{$refA_hosts}){
		my ($session, $error) = Net::SNMP->session(	-version	=> 'snmpv2c',
								-hostname   	=> $ip,
								-community   	=> $self->{_comunity},
#								-translate	=> [-octetstring => 0],
								-retries	=> $self->{_retries},
								-timeout	=> $self->{_timeout},
								-nonblocking	=> 1,
							);

		if (!defined($session)) {
			print "ERROR: SNMP initialisation\n";
			$self->{_ERRMSG} = $error;
			die;
		}

		my $result = $session->$action(%params);
		
		if (!defined($result)) {
			my $msg = sprintf "Failed to queue get request for host '%s': %s.", $session->hostname(), $session->error();
			print "ERROR: $msg";
			$self->{_logger}->error($msg);
			$self->{_ERRMSG} = $error;
			$session->close;
		}
	}
	
	my $processTime = sprintf "%.3f", (gettimeofday - $startTime) / 1000;
	$self->{_logger}->debug("waiting for SNMP responses ($processTime ms)");
	Net::SNMP::snmp_dispatcher();
	$startTime = 0;
	
	$self->removeBadIp($refA_hosts);
}

#######################################################################################
#determineOID()
sub determineOID {
	my $self	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my @MIBs	= ($sysDescr, $sysObjectID, $sysName);
	
	$self->{_logger}->debug("run 'system info' for ".@{$refA_hosts}." hosts");
	$self->performSNMP('get_request', 'determineOID', $refA_hosts, \@MIBs, $refO_data);
}

#######################################################################################
#determineOID_cb()
sub determineOID_cb {
	my $session	= shift;
	my $logger	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $result	= $session->var_bind_list();
	my $ip		= $session->hostname;
	my $processTime = sprintf "%.3f", (gettimeofday - $startTime) * 1000;
	
  	if($session->{'_error'}){
		$$refO_data{$ip}->setValue('Status', $session->{'_error'});
		markBadIp($refA_hosts, $ip);
		$logger->warn("no response form host $ip");
  	}elsif($result->{$sysObjectID} !~ /^1\.3\.6\.1\.4\.1\.45\.3\./ or
	       $result->{$sysObjectID} =~ /^1\.3\.6\.1\.4\.1\.45\.3\.79\./){
		$$refO_data{$ip}->setValue('Status', "don't support AES");
		markBadIp($refA_hosts, $ip);
		$logger->info("host $ip is not AES capable");
	}else{
		$$refO_data{$ip}->setValue('OID', $result->{$sysObjectID});
		$$refO_data{$ip}->setValue('Name', $result->{$sysName});
		$$refO_data{$ip}->setValue('Description', $result->{$sysDescr});
		$result->{$sysDescr} =~ /SW\:v(\d.\d.\d.\d*) /;
		$$refO_data{$ip}->setValue('Software', $1);
		$result->{$sysDescr} =~ /\s(\d[^\s]*)\s/;
		$$refO_data{$ip}->setValue('Type', 'ERS-'.$1);
		$$refO_data{$ip}->setValue('Status', 'up');
		
		$logger->debug("got 'system info' for host $ip ($processTime)");
  	}
}

#######################################################################################
#getEnergySaverSchedule()
sub getEnergySaverSchedule {
	my $self	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my @MIBs	= ($bsnesScheduleAction);
	
	$self->{_logger}->debug("run 'AES schedule' for ".@{$refA_hosts}." hosts");
	$self->performSNMP('get_bulk_request', 'getEnergySaverSchedule', $refA_hosts, \@MIBs, $refO_data);
}

#######################################################################################
#getEnergySaverSchedule_cb()
sub getEnergySaverSchedule_cb {
	my $session	= shift;
	my $logger	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $result	= $session->var_bind_list();
	my $ip		= $session->hostname;
	my @action	= qw(dummy activate deactivate);
	my $pattern	= "^".$bsnesScheduleAction.'\.(\d)\.(\d+).(\d+)';
	my @record;
	my $processTime = sprintf "%.3f", (gettimeofday - $startTime) * 1000;
	
	if($session->{'_error'}){
  		markBadIp($refA_hosts, $ip);
		$logger->warn("got 'AES schedule' for host $ip fails, '".$session->{'_error'});
	}else{
		foreach my $mib (sort keys %{$result}){
			my %data;
			if($mib =~ /$pattern/){
				$data{'day'} = $1;
				$data{'hour'} = $2;
				$data{'hour'} = '0'.$data{'hour'} if (length($data{'hour'}) <2);
				$data{'minute'} = $3;
				$data{'minute'} = '0'.$data{'minute'} if (length($data{'minute'}) <2);
				$data{'action'} = $action[$result->{$mib}];
				$logger->trace("got 'AES schedule' record '$data{'day'} $data{'hour'}:$data{'minute'} = $result->{$mib}' for host $ip");
				push(@record, \%data);
			}
		}
		$$refO_data{$ip}->setValue('EnergySaverSchedule', \@record);
	}
	
	my $recCount = @record || 0;
	$logger->debug("got 'AES schedule' [$recCount] for host $ip ($processTime)");
}

#######################################################################################
#getUnitCount()
sub getUnitCount {
	my $self	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my @MIBs	= ($s5ChasComDescr);
	
	$self->{_logger}->debug("run 'unit count' for ".@{$refA_hosts}." hosts");
	$self->performSNMP('get_bulk_request', 'getUnitCount', $refA_hosts, \@MIBs, $refO_data);
}

#######################################################################################
#getUnitCount_cb()
sub getUnitCount_cb {
	my $session	= shift;
	my $logger	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $result	= $session->var_bind_list();
	my $ip		= $session->hostname;
	my $processTime = sprintf "%.3f", (gettimeofday - $startTime) * 1000;
	
	if($session->{'_error'}){
  		markBadIp($refA_hosts, $ip);
		$logger->warn("got 'unit count' for host $ip fails, '".$session->{'_error'});
	}else{
		my $count = 0;
		foreach my $mib (keys %{$result}){
			next if $mib !~ /$s5ChasComDescr/;
			$count++
		}
		$$refO_data{$ip}->setValue('UnitCount', $count);
		$logger->debug("got 'unit count' [$count] for host $ip ($processTime)");
	}
}

#######################################################################################
#getPortCount()
sub getPortCount {
	my $self	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my @MIBs	= ($ifNumber);
	
	$self->{_logger}->debug("run 'port count' for ".@{$refA_hosts}." hosts");
	$self->performSNMP('get_request', 'getPortCount', $refA_hosts, \@MIBs, $refO_data);
}

#######################################################################################
#getPortCount_cb()
sub getPortCount_cb {
	my $session	= shift;
	my $logger	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $result	= $session->var_bind_list();
	my $ip		= $session->hostname;
	my $processTime = sprintf "%.3f", (gettimeofday - $startTime) * 1000;
	
	if($session->{'_error'}){
  		markBadIp($refA_hosts, $ip);
		$logger->warn("got 'port count' for host $ip fails, ".$session->{'_error'});
	}else{
		$$refO_data{$ip}->setValue('PortCount', $result->{$ifNumber});
		$logger->debug("got 'port count' [".$result->{$ifNumber}."] for host $ip ($processTime)");
	}
}

#######################################################################################
#getEnergySaverEnabled()
sub getEnergySaverEnabled {
	my $self	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my @MIBs	= ($bsnesEnergySaverEnabled);
	
	$self->{_logger}->debug("run 'AES global' for ".@{$refA_hosts}." hosts");
	$self->performSNMP('get_request', 'getEnergySaverEnabled', $refA_hosts, \@MIBs, $refO_data);
}

#######################################################################################
#getEnergySaverEnabled_cb()
sub getEnergySaverEnabled_cb {
	my $session	= shift;
	my $logger	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $result	= $session->var_bind_list();
	my $ip		= $session->hostname;
	my @result 	= qw(dummy enabled disabled);
	my $processTime = sprintf "%.3f", (gettimeofday - $startTime) * 1000;
	
	if($session->{'_error'}){
  		markBadIp($refA_hosts, $ip);
		$logger->warn("got 'AES global' for host $ip fails, '".$session->{'_error'});
	}else{
		if($result->{$bsnesEnergySaverEnabled} == 2){
			$$refO_data{$ip}->setValue('EnergySaverEnabled', 'disabled');
		}else{
			$$refO_data{$ip}->setValue('EnergySaverEnabled', 'enabled');
		}
		
		$logger->debug("got 'AES global' [".$result[$result->{$bsnesEnergySaverEnabled}]."(".$result->{$bsnesEnergySaverEnabled}.")] for host $ip ($processTime)");
	}
}

#######################################################################################
#interfaceEnergySaverEnabled()
sub interfaceEnergySaverEnabled {
	my $self	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $maxRepition = $maxBulkRepition;
	my @hosts	= @{$refA_hosts};
	my $lastIndex	= 0;
	my $loop	= 1;
	
	do{
		$self->{_logger}->debug("run 'AES interface' [loop($loop)/index($lastIndex)] for ".@hosts." hosts");
		my @MIBs = ($bsnesInterfaceEnergySaverEnabled.'.'.$lastIndex);
		$self->performSNMP('get_bulk_request', 'interfaceEnergySaverEnabled', \@hosts, \@MIBs, $refO_data, $maxRepition);
		
		for(my $count = 0; $count < @hosts; $count++){
			if($$refO_data{$hosts[$count]}->getValue('lastIndex') >= $$refO_data{$hosts[$count]}->getValue('PortCount')){
				$self->{_logger}->trace("remove host ".$$refO_data{$hosts[$count]}->getValue('_IP')." form list, last index reached");
				splice(@hosts, $count, 1);
				$count--;
			}
		}
		$lastIndex = $lastIndex + $maxRepition;
		$loop++;
	}while(@hosts);
	
	$self->resetLastIndex($refA_hosts, $refO_data);
}

#######################################################################################
#interfaceEnergySaverEnabled_cb()
sub interfaceEnergySaverEnabled_cb {
	my $session	= shift;
	my $logger	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $result	= $session->var_bind_list();
	my $ip		= $session->hostname;
	my $lastIndex	= 0;
	my $countEnabled= 0;
	my $pattern	= '^'.$bsnesInterfaceEnergySaverEnabled;
	my $processTime = sprintf "%.3f", (gettimeofday - $startTime) * 1000;
	
	if($session->{'_error'}){
  		markBadIp($refA_hosts, $ip);
		$logger->warn("got 'AES interface' for host $ip fails, '".$session->{'_error'});
	}else{
#		if($$refO_data{$ip}->getValue('interfaceEnergySaverEnabled') eq 'X'){
#			$$refO_data{$ip}->setValue('interfaceEnergySaverEnabled', 0);
#		}
		foreach my $mib (sort keys %{$result}){
			if($mib !~ /$pattern/){
				$lastIndex = 100000;
				last;
			}
			if($result->{$mib} eq 1){
				$countEnabled++;
			}
			$mib =~ /\.(\d+)$/;
			$lastIndex = $1 if $lastIndex < $1;
		}
	}
	
	$lastIndex = 100000 if !$lastIndex;
	
	$$refO_data{$ip}->setValue('lastIndex', $lastIndex);
	$countEnabled = $countEnabled + $$refO_data{$ip}->getValue('interfaceEnergySaverEnabled');
	$$refO_data{$ip}->setValue('interfaceEnergySaverEnabled', $countEnabled);
	
	if($lastIndex == 100000){
		$logger->debug("got 'AES interface' [$countEnabled] for host $ip (finished) ($processTime)");
	}else{
		$logger->debug("got 'AES interface' [$countEnabled] for host $ip (index:$lastIndex) ($processTime)");
	}
}

#######################################################################################
#energyUnitSavings()
sub energyUnitSavings {
	my $self	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $maxRepition = $maxBulkRepition;
	my @hosts	= @{$refA_hosts};
	my $lastIndex	= 0;
	my $loop	= 1;
	
	do{
		$self->{_logger}->debug("run 'AES savings' [$loop/$lastIndex] for ".@hosts." hosts");
		my @MIBs = ($bsnesSavingsUnitSavings.'.'.$lastIndex);
		$self->performSNMP('get_bulk_request', 'energyUnitSavings', \@hosts, \@MIBs, $refO_data, $maxRepition);
		
		for(my $count = 0; $count < @hosts; $count++){
			if($$refO_data{$hosts[$count]}->getValue('lastIndex') >= $$refO_data{$hosts[$count]}->getValue('PortCount')){
				$self->{_logger}->trace("remove host ".$$refO_data{$hosts[$count]}->getValue('_IP')." form list, last index reached");
				splice(@hosts, $count, 1);
				$count--;
			}
		}
		$lastIndex = $lastIndex + $maxRepition;
		$loop++;
	}while(@hosts);
	
	$self->resetLastIndex($refA_hosts, $refO_data);
}

#######################################################################################
#energyUnitSavings_cb()
sub energyUnitSavings_cb {
	my $session	= shift;
	my $logger	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	my $result	= $session->var_bind_list();
	my $ip		= $session->hostname;
	my $lastIndex	= 0;
	my $unitCount	= 0;
	my $pattern	= '^'.$bsnesSavingsUnitSavings;
	my $processTime = sprintf "%.3f", (gettimeofday - $startTime) * 1000;
	
	if($session->{'_error'}){
  		markBadIp($refA_hosts, $ip);
		$logger->warn("got 'AES savings' for host $ip fails, '".$session->{'_error'});
	}else{
#		if($$refO_data{$ip}->getValue('EnergyUnitSavings') eq 'X'){
#			$$refO_data{$ip}->setValue('EnergyUnitSavings', 0);
#		}
		foreach my $mib (sort keys %{$result}){
			if($mib !~ /$pattern/){
				$lastIndex = 100000;
				last;
			}
			$unitCount = $unitCount + $result->{$mib};
			$mib =~ /\.(\d+)$/;
			$lastIndex = $1 if $lastIndex < $1;
		}
	}
	
	$lastIndex = 100000 if !$lastIndex;
	
	$$refO_data{$ip}->setValue('lastIndex', $lastIndex);
	$unitCount = $$refO_data{$ip}->getValue('EnergyUnitSavings') + $unitCount / 10;
	$$refO_data{$ip}->setValue('EnergyUnitSavings', $unitCount);
	
	if($lastIndex == 100000){
		$logger->debug("got 'AES savings' [$unitCount] for host $ip (finished) ($processTime)");
	}else{
		$logger->debug("got 'AES savings' [$unitCount] for host $ip (index:$lastIndex) ($processTime)");
	}
}

#######################################################################################
#resetLastIndex()
sub resetLastIndex {
	my $self	= shift;
	my $refA_hosts	= shift;
	my $refO_data	= shift;
	
	foreach my $ip (@{$refA_hosts}){
		$$refO_data{$ip}->setValue('lastIndex', 0);
	}
	
	$self->{_logger}->trace("reset SNMP index");
}

#######################################################################################
#markBadIp()
sub markBadIp {
	my $refA_hosts	= shift;
	my $badIp = shift;
	
	for(my $count = 0; $count < @{$refA_hosts}; $count++){
		if($$refA_hosts[$count] eq $badIp){
			$$refA_hosts[$count] = 0;
			return;
		}
	}
}

#######################################################################################
#removeBadIp()
sub removeBadIp {
	my $self	= shift;
	my $refA_hosts	= shift;
	my $count	= 0;
	
	for(my $index = 0; $index < @{$refA_hosts}; $index++){
		if(!$$refA_hosts[$index]){
			splice(@{$refA_hosts}, $index, 1);
			$index--;
			$count++;
		}
	}
	
	$self->{_logger}->debug("remove $count bad hosts from processing list") if $count;
}

1;
